"""
Copyright (c) 2020 Lumerical Inc.
This program is commercial software: you can use it under the terms of the
Lumerical License Agreement as published by Lumerical Inc. You should have
received a copy of the Lumerical License Agreement along with this program.
If not, see:
https://www.lumerical.com/tcad-products/licensing/license_agreement.html.

Except as expressly permitted in the Lumerical License Agreement, you may not
modify or redistribute this program.

2021-11: Edited by Amnon Wilinger for PBS queue 
"""

import traceback
import subprocess
import sys
import io
import platform
from time import sleep
from os.path import split, splitext, expanduser, abspath, join
from pathlib import PurePosixPath
import re
from datetime import datetime


USE_SSH = True
USE_SCP = True
CLUSTER_CWD = ''
if USE_SSH:
    SSH_LOGIN = f"amnonjw@zeus.technion.ac.il"
    now = datetime.now()
    dt_string = now.strftime("%Y-%m-%d--%H-%M-%S")
    CLUSTER_CWD = "$HOME/Lumerical/remote/" + dt_string + "/"  # NFS directory shared by all nodes, can be '$HOME/' or a full path, ending in '/'
if not USE_SCP:
    PATH_TRANSLATION = ('','') # Default, no translation. paths are 1:1 match. Use this in the case that the GUI is running on the cluster already.
    # PATH_TRANSLATION = ("X:/Working Dir/", "/share/working/dir/") # use unix-style path deliminters '/'. Use this in the case that there is a shared file system on a windows machine to a linux cluster, or if the host machine has different mount points for the shared directory.

def log(msg):
    t = datetime.now().isoformat()
    print(t + '\t' + msg)
    sys.stdout.flush()


def posix_path(p):
    return str(PurePosixPath(p))


def call(cmd):
    if platform.system() == "Windows":  # prevent cmd windows from popping up
        si = subprocess.STARTUPINFO()
        si.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        subprocess.call(cmd, startupinfo=si)
    else: 
        subprocess.call(cmd)


def check_output(cmd):
    if platform.system() == "Windows":  # prevent cmd windows from popping up
        si = subprocess.STARTUPINFO()
        si.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        return subprocess.check_output(cmd, startupinfo=si).strip().decode()
    else: 
        return subprocess.check_output(cmd).strip().decode()


def ssh_wrapper(arr):
    cmd = ''
    if(CLUSTER_CWD):
        cmd += f'cd {CLUSTER_CWD};' # if the directory does not exist this will fail silently. mk_remote_dir must be called explicitly.
    cmd += ' '.join(arr)
    cmd = ['ssh', SSH_LOGIN, cmd]
    return cmd


def mk_remote_dir(d=CLUSTER_CWD):
    assert(d)
    cmd = ssh_wrapper(['mkdir', '-p', d])

    log(f'Creating remote dir: {d}. Calling command: {" ".join(cmd)}') # 
    call(cmd)

def remote_clean(fsp_basename):
    cmd = ssh_wrapper(['rm', CLUSTER_CWD + fsp_basename + '.o*'])    
    log(f'Cleaning remote files. Calling command: {" ".join(cmd)}')
    call(cmd)
    cmd = ssh_wrapper(['rm', CLUSTER_CWD + fsp_basename + '.e*'])    
    log(f'Cleaning remote files. Calling command: {" ".join(cmd)}')
    call(cmd)

def put(fsp_file):
    cmd = ['scp', fsp_file, SSH_LOGIN + ':' + CLUSTER_CWD]
    log(f'Copying local file "{fsp_file}" to remote host. Calling command: {" ".join(cmd)}')
    call(cmd)

def get(fsp_basename):
    cmd = ['scp', SSH_LOGIN + ':' + CLUSTER_CWD + fsp_basename + '.fsp', '.']
    log(f'Copying remote file to local host. Calling command {" ".join(cmd)}')
    call(cmd)

def get_log_file(fsp_basename):
    cmd = ['scp', SSH_LOGIN + ':' + CLUSTER_CWD + fsp_basename + '_p0.log', '.']    
    try:
        call(cmd)
        file_handle = io.open(fsp_basename + '_p0.log',"r")
        lines = file_handle.readlines()
    except Exception as ex:
        log(f'Copying remote file to local host. Calling command {" ".join(cmd)}')
        log(traceback.format_exc())
        lines = []
    return lines

status_line_marker = "complete. Max time remaining"
def update_progress_from_log_file(log_lines, last_line):
    if (len(log_lines)-1 != last_line):  # there was a change in the log content
        for i_line in range(last_line, len(log_lines)):
            if status_line_marker in log_lines[i_line]:
                log(log_lines[i_line])
                print(log_lines[i_line])
        sys.stdout.flush()
        last_line = i_line
    return last_line

def remote_path_substitution(local_path):
    filepath, filename = split(local_path)
    remote_path = None
    if USE_SCP:
        remote_path = CLUSTER_CWD + filename # SCP defults to $HOME
    else:
        remote_path = join(filepath.replace(PATH_TRANSLATION[0], PATH_TRANSLATION[1]), filename) # abspath forces unix-style delimiters
    remote_path = posix_path(remote_path)
    log(f'Translated local path "{local_path}" to remote path "{remote_path}"')
    return remote_path


def parse_submission_script(submission_script_lines):
# expects a quoted path for a file with extension ['.fsp', '.icp', '.lms', '.ldev']
    local_path = ''
    filename = ''
    basename = ''
    for i in range(len(submission_script_lines)):
        line = submission_script_lines[i]
        # filepath must be a double-quoted string
        quoted_args = re.findall('"([^"]*)"', line)
        if (len(quoted_args) > 0):
            for arg in quoted_args:
                if any(arg.endswith(file_extension) for file_extension in ['.fsp', '.icp', '.lms', '.ldev']):
                    local_path = arg
                    filename = split(local_path)[1]
                    basename = splitext(filename)[0]
                    submission_script_lines[i] = line.replace(local_path, remote_path_substitution(local_path))
                    break

    if not basename:
            raise Exception("A project file (.fsp, .ipc, .lms, .ldev) was not found in the provided arguments: {}".format(submission_script))
    assert(local_path != '')
    assert(filename != '')
    assert(basename != '')
    return submission_script_lines, local_path, filename, basename


def run_job(submission_script_lines, submission_command):
    assert((len(submission_command) > 0))

    submission_script_lines, local_path, filename, basename = parse_submission_script(submission_script_lines)

    if USE_SSH and USE_SCP:
        put(local_path)

    job_id = submit_job(submission_script_lines, submission_command, basename)
    sleep(0.5)
    assert(job_in_queue(job_id))
    print("<status>QUEUED</status>")
    sys.stdout.flush()

    while job_status(job_id,basename) == 'Q':
        sleep(1)
    if job_status(job_id,basename) == 'R':
        print("<status>RUNNING</status>")
        sleep(2) # wait before attempting to retrive the log file
        last_log_line = -1 # the last line read in previous reading of te log file
        while job_status(job_id,basename) != 'C':
            log_lines = get_log_file(basename)
            last_log_line = update_progress_from_log_file(log_lines,last_log_line)
            sleep(1)

    if USE_SSH and USE_SCP:
        remote_clean(basename)
        get(basename)


def kill_job(job_id):
    cmd = ['qdel', job_id]
    if USE_SSH:
        cmd = ssh_wrapper(cmd)
    log("Kill job: " + ' '.join(cmd))
    call(cmd)


def submit_job(submission_script, submission_command, job_name):
    cmd = submission_command + ['-N', job_name]
    if USE_SSH:
        cmd = ssh_wrapper(cmd)
        
    log('Submission Command: ' + ' '.join(cmd))
    log('Submission Script:\n' + '\n'.join(submission_script))
    p = subprocess.Popen(cmd, shell=True,stdin=subprocess.PIPE,stdout=subprocess.PIPE, encoding='utf8', universal_newlines=True)
    p.stdin.buffer.write('\n'.join(submission_script).encode()) # ensure unix style line endings are used
    p.stdin.close()

    result = p.stdout.readlines()
    log('Submission Response:\n' + '\n'.join(result))

    result = result[0].split('.')
    job_id = result[0]
    assert(job_id.isdigit())
    log(f'Submission successful, Job ID: {job_id}')
    return job_id


def job_in_queue(job_id):
    cmd = ['qstat', job_id]
    if USE_SSH:
        cmd = ssh_wrapper(cmd)
    job_status = check_output(cmd)
    for line in job_status.splitlines():
        possible_job_id = line.strip().split()[0]
        possible_job_id = possible_job_id.split('.')[0]
        if possible_job_id == str(job_id):
            return True
    return False

def job_status(job_id,basename):
    cmd = ['qstat', job_id]
    if USE_SSH:
        cmd = ssh_wrapper(cmd)
    try:
        job_status = check_output(cmd)
        for line in job_status.splitlines():
            possible_job_id = line.strip().split()[0]
            possible_job_id = possible_job_id.split('.')[0]
            if possible_job_id == str(job_id):
                status = line.strip().split()[4]
                sys.stdout.flush()
                return status
    except subprocess.CalledProcessError as e:
        print("python subprocess error - possible job finished and popped out of queue table")
        status  = 'C' 
        print("<simComplete/>")
        print("<complete/>")
        return status


if __name__ == '__main__':
    submission_command = sys.argv[1:]
    submission_script = sys.stdin.read().splitlines()
    if(CLUSTER_CWD):
        mk_remote_dir()
    run_job(submission_script, submission_command)
